import opt_einsum
import torch


def similarity_vec(Z1, Z2=None):  # (M, D), (N, D)
    """
    Takes the pairwise inner products of all rows of Z1 and Z2.

    Args:
        Z1: first data matrix (M, D)
        Z2: second data matrix (N, D)

    Returns:
        A matrix containing the pairwise inner products (M, N)
    """
    if Z2 is None:
        Z2 = Z1
    return Z1.conj() @ Z2.T  # (M, N)


def similarity_shift_invariant(Z1, Z2=None):  # (M, C, T), (N, C, T)
    """
    Takes the pairwise inner products of all rows of Z1 and Z2.

    Args:
        Z1: first data matrix (M, C, T)
        Z2: second data matrix (N, C, T)

    Returns:
        A matrix containing the pairwise inner products (M, N)
    """
    if Z2 is None:
        Z2 = Z1
    return opt_einsum.contract("mct, nct -> mn", Z1.conj(), Z2)  # (M, N)


def similarity_translation_invariant(Z1, Z2=None):  # (M, C, H, W), (N, C, H, W)
    """
    Takes the pairwise inner products of all rows of Z1 and Z2.

    Args:
        Z1: first data matrix (M, C, H, W)
        Z2: second data matrix (N, C, H, W)

    Returns:
        A matrix containing the pairwise inner products (M, N)
    """
    if Z2 is None:
        Z2 = Z1
    return opt_einsum.contract("mchw, nchw -> mn", Z1.conj(), Z2)  # (M, N)


def Sigma_hat_vec(Z):  # (N, D)
    """
    Computes the empirical covariance of a data matrix Z.
    Args:
        Z: a data matrix (N, D)

    Returns:
        The covariance matrix of Z data (D, D)
    """
    return Z.T.conj() @ Z / Z.shape[0]  # (N, D) x (N, D) -> (D, D)


def Sigma_hat_per_class_vec(Z, Pi):  # (N, D), (N, K)
    """
    Computes the empirical covariance  of each class of a data matrix Z.
    Args:
        Z: a data matrix (N, D)
        Pi: a class information matrix (N, K)

    Returns:
        The class covariance tensor of Z data (K, D, D)
    """
    K = Pi.shape[1]
    N_per_class = torch.sum(Pi, axis=0)  # (K, )
    weight_per_class = torch.where(  # stops divide by 0 errors
        N_per_class > 0,
        1 / N_per_class,  # (K, )
        torch.tensor(0.0)  # ()
    )  # (K, )
    ZPiZh = opt_einsum.contract("ni, nj, nk -> kij", Z.conj(), Z,
                                Pi.to(torch.float))  # (N, D) x (N, D) x (K, N) -> (K, D, D)
    return weight_per_class.view(K, 1, 1) * ZPiZh


def Sigma_hat_shift_invariant(Z):  # (N, C, T)
    """
    Computes the empirical covariance of a multi-channel 1D data matrix Z.
    Should be used in contexts where Z matrix is already FFTed in last dimension.
    Args:
        Z: a data matrix (N, C, T)

    Returns:
        The covariance matrix of Z data (T, C, C)
    """
    return opt_einsum.contract("nct, ndt -> tcd", Z.conj(), Z) / Z.shape[0]  # (N, C, T) x (N, C, T) ->  (T, C, C)


def Sigma_hat_per_class_shift_invariant(Z, Pi):  # (K, T, C, C),  (N, K)
    """
    Computes the empirical covariance of each class of a multi-channel 1D data matrix Z.
    Should be used in contexts where Z matrix is already FFTed in last dimension.
    Args:
        Z: a data matrix (N, C, T)
        Pi: a class information matrix (N, K)

    Returns:
        The class covariance tensor of Z data (K, T, C, C)
    """
    K = Pi.shape[1]
    N_per_class = torch.sum(Pi, axis=0)  # (K, )
    weight_per_class = torch.where(  # stops divide by 0 errors
        N_per_class > 0,
        1 / N_per_class,  # (K, )
        torch.tensor(0.0)  # ()
    )  # (K, )
    ZPiZh = opt_einsum.contract("nct, ndt, nk -> ktcd", Z.conj(), Z,
                                Pi.to(torch.cfloat))  # (N, C, T) x (N, C, T) x (K, N) -> (K, T, C, C)
    return weight_per_class.view(K, 1, 1, 1) * ZPiZh


def Sigma_hat_translation_invariant(Z):  # (N, C, H, W)
    """
    Computes the empirical covariance of a multi-channel 2D data matrix Z.
    Should be used in contexts where Z matrix is already FFTed in last two dimensions.
    Args:
        Z: a data matrix (N, C, H, W)

    Returns:
        The covariance matrix of Z data (H, W, C, C)
    """
    return opt_einsum.contract("nchw, ndhw -> hwcd", Z.conj(), Z) / Z.shape[
        0]  # (N, C, H, W) x (N, C, H, W) -> (H, W, C, C)


def Sigma_hat_per_class_translation_invariant(Z, Pi):  # (N, C, H, W), (N, K)
    """
    Computes the empirical covariance of each class of a multi-channel 2D data matrix Z.
    Should be used in contexts where Z matrix is already FFTed in last two dimensions.
    Args:
        Z: a data matrix (N, C, H, W)
        Pi: a class information matrix (N, K)

    Returns:
        The class covariance tensor of Z data (K, H, W, C, C)
    """
    K = Pi.shape[1]
    N_per_class = torch.sum(Pi, axis=0)  # (K, )
    weight_per_class = torch.where(  # stops divide by 0 errors
        N_per_class > 0,
        1 / N_per_class,  # (K, )
        torch.tensor(0.0)  # ()
    )  # (K, )
    ZPiZh = opt_einsum.contract("nchw, ndhw, nk -> khwcd", Z.conj(), Z,
                                Pi.to(torch.cfloat))  # (N, C, H, W) x (N, C, H, W) x (N, K) -> (K, H, W, C, C)
    return weight_per_class.view(K, 1, 1, 1, 1) * ZPiZh


__all__ = [
    "similarity_vec", "similarity_shift_invariant", "similarity_translation_invariant",
    "Sigma_hat_vec", "Sigma_hat_per_class_vec",
    "Sigma_hat_shift_invariant", "Sigma_hat_per_class_shift_invariant",
    "Sigma_hat_translation_invariant", "Sigma_hat_per_class_translation_invariant"
]



import mcr2
import torch


class SupervisedLDRLoss(torch.nn.Module):
    def __init__(self, eps_sq):
        super(SupervisedLDRLoss, self).__init__()
        self.cr = mcr2.coding_rate.SupervisedVectorCodingRate(eps_sq)


class SupervisedBinaryLDRLoss(SupervisedLDRLoss):
    def forward(self, Z, Z_hat):
        return self.cr.DeltaR_distance(Z, Z_hat)


class SupervisedMulticlassLDRLoss(SupervisedLDRLoss):
    def forward(self, Z, Z_hat, Pi):
        N = Z.shape[0]
        N_per_class = torch.sum(Pi, axis=0)
        gamma_per_class = N_per_class / N

        R_Z = self.cr.R(Z)  # ()
        R_Zhat = self.cr.R(Z_hat)  # ()
        R_Z_per_class = self.cr.R_per_class(Z, Pi)  # (K, )
        R_Zhat_per_class = self.cr.R_per_class(Z_hat, Pi)  # (K, )

        R_Zi_Zhati = sum(self.cr.R(
            torch.cat(tensors=(Z[Pi[:, i] == 1], Z_hat[Pi[:, i] == 1]), dim=0)
        ) for i in range(Pi.shape[1]))  # TODO: optimize out CPU access

        return R_Z + R_Zhat + R_Zi_Zhati - torch.sum((gamma_per_class + 0.5) * (R_Z_per_class + R_Zhat_per_class))  # ()


__all__ = ["SupervisedBinaryLDRLoss", "SupervisedMulticlassLDRLoss"]



import mcr2.functional as F
import torch


class SupervisedCodingRate:
    def __init__(self, eps_sq):
        self.eps_sq = eps_sq

    def R(self, Z):
        raise NotImplementedError("Called R in basic coding rate class")

    def Rc(self, Z, Pi):
        raise NotImplementedError("Called Rc in basic coding rate class")

    def DeltaR(self, Z, Pi):
        return self.R(Z) - self.Rc(Z, Pi)

    def DeltaR_distance(self, Z1, Z2):
        N1 = Z1.shape[0]
        N2 = Z2.shape[0]
        N = N1 + N2
        return self.R(torch.cat(tensors=(Z1, Z2), dim=0)) - (N1 / N) * self.R(Z1) - (N2 / N) * self.R(Z2)


class SupervisedVectorCodingRate(SupervisedCodingRate):
    def R(self, Z):
        """
        Computes the coding rate of Z.

        Args:
            Z: data matrix, (N, D)

        Returns:
            The coding rate of Z, ().
        """
        N, D = Z.shape  # (N, D)
        Sigma_hat_data = F.Sigma_hat_vec(Z)  # (D, D)
        beta = D / self.eps_sq  # ()
        I = torch.eye(D)  # (D, D)
        Sigma_hat_distorted = I + beta * Sigma_hat_data  # (D, D)
        return F.logdet(Sigma_hat_distorted) / 2.0  # ()

    def R_per_class(self, Z, Pi):
        """
        Computes the coding rate of each class of Z as given by class information matrix Pi.

        Args:
            Z: data matrix, (N, D)
            Pi: class matrix, (N, K)

        Returns:
            The coding rate of each class of Z, (K, ).
        """
        N, D = Z.shape  # (N, D)
        Sigma_hat_data_per_class = F.Sigma_hat_per_class_vec(Z, Pi)  # (K, D, D)
        beta = D / self.eps_sq  # ()
        I = torch.eye(D).view(1, D, D)  # (D, D)
        Sigma_hat_distorted_per_class = I + beta * Sigma_hat_data_per_class  # (K, D, D)
        logdets_per_class = F.logdet(Sigma_hat_distorted_per_class)  # (K, )
        return logdets_per_class / 2.0  # ()

    def Rc(self, Z, Pi):
        """
        Computes the segmented coding rate of Z with respect to a class information matrix Pi.

        Args:
            Z: data matrix, (N, D)
            Pi: class information matrix, (N, K)

        Returns:
            The segmented coding rate of Z with respect to Pi, ().
        """
        N, K = Pi.shape  # (N, K)
        N_per_class = torch.sum(Pi, axis=0)  # (K, )
        gamma_per_class = N_per_class / N  # (K, )
        Rs_per_class = self.R_per_class(Z, Pi)  # (K, )
        return torch.sum(gamma_per_class * Rs_per_class)  # ()

    def DeltaR(self, Z, Pi):
        """
        Computes the coding rate reduction of Z with respect to a class information matrix Pi.

        Args:
            Z: data matrix, (N, D)
            Pi: class information matrix, (K, N)

        Returns:
            The coding rate reduction of Z with respect to Pi, ().
        """
        return super().DeltaR(Z, Pi)

    def DeltaR_distance(self, Z1, Z2):
        """
        Computes the DeltaR distance between the two data point sets Z1 and Z2.

        Args:
            Z1: data matrix, (N1, D)
            Z2: data matrix, (N2, D)

        Returns:
            The DeltaR distance between Z1 and Z2, ().
        """
        return super().DeltaR_distance(Z1, Z2)


class SupervisedShiftInvariantCodingRate(SupervisedCodingRate):
    def R(self, Z):
        """
        Computes the coding rate of Z.
        NOTE: Only an accurate measure of the coding rate if the data is FFTed beforehand.
        You can do this with mcr2.functional.fft(Z).

        Args:
            Z: data matrix, (N, C, T)

        Returns:
            The coding rate of Z, ().
        """
        N, C, T = Z.shape  # (N, C, T)
        Sigma_hat_data = F.Sigma_hat_shift_invariant(Z)  # (T, C, C)
        beta = C / self.eps_sq  # ()
        I = torch.eye(C).view(1, C, C)  # (1, C, C)
        Sigma_hat_distorted = I + beta * Sigma_hat_data  # (T, C, C)
        return torch.sum(F.logdet_complex(Sigma_hat_distorted)) / 2.0  # ()

    def R_per_class(self, Z, Pi):
        """
        Computes the coding rate of each class of Z as given by class information matrix Pi.
        NOTE: Only an accurate measure of the coding rate if the data is FFTed beforehand.
        You can do this with mcr2.functional.fft(Z).

        Args:
            Z: data matrix, (N, C, T)
            Pi: class matrix, (N, K)

        Returns:
            The coding rate of each class of Z, (K, ).
        """
        N, C, T = Z.shape  # (N, C, T)
        Sigma_hat_data_per_class = F.Sigma_hat_per_class_shift_invariant(Z, Pi)  # (K, T, C, C)
        beta = C / self.eps_sq  # ()
        I = torch.eye(C).view(1, 1, C, C)  # (1, 1, C, C)
        Sigma_hat_distorted_per_class = I + beta * Sigma_hat_data_per_class  # (K, T, C, C)
        logdets_per_class = torch.sum(
            F.logdet_complex(Sigma_hat_distorted_per_class),  # (K, T)
            dim=1
        )  # (K, )
        return logdets_per_class / 2.0  # ()

    def Rc(self, Z, Pi):
        """
        Computes the segmented coding rate of Z with respect to a class information matrix Pi.
        NOTE: Only an accurate measure of the coding rate if the data is FFTed beforehand.
        You can do this with mcr2.functional.fft(Z).

        Args:
            Z: data matrix, (N, C, T)
            Pi: class information matrix, (N, K)

        Returns:
            The segmented coding rate of Z with respect to Pi, ().
        """
        N, K = Pi.shape  # (N, K)
        N_per_class = torch.sum(Pi, axis=0)  # (K, )
        gamma_per_class = N_per_class / N  # (K, )
        Rs_per_class = self.R_per_class(Z, Pi)  # (K, )
        return torch.sum(gamma_per_class * Rs_per_class)  # ()

    def DeltaR(self, Z, Pi):
        """
        Computes the coding rate reduction of Z with respect to a class information matrix Pi.
        NOTE: Only an accurate measure of the coding rate if the data is FFTed beforehand.
        You can do this with mcr2.functional.fft(Z).

        Args:
            Z: data matrix, (N, C, T)
            Pi: class information matrix, (N, K)

        Returns:
            The coding rate reduction of Z with respect to Pi, ().
        """
        return super().DeltaR(Z, Pi)

    def DeltaR_distance(self, Z1, Z2):
        """
        Computes the DeltaR distance between the two data point sets Z1 and Z2.
        NOTE: Only an accurate measure of the distance if the data is FFTed beforehand.
        You can do this with mcr2.functional.fft(Z).

        Args:
            Z1: data matrix, (N1, C, T)
            Z2: data matrix, (N2, C, T)

        Returns:
            The DeltaR distance between Z1 and Z2, ().
        """
        return super().DeltaR_distance(Z1, Z2)


class SupervisedTranslationInvariantCodingRate(SupervisedCodingRate):
    def R(self, Z):
        """
        Computes the coding rate of Z.
        NOTE: Only an accurate measure of the coding rate if the data is FFTed beforehand.
        You can do this with mcr2.functional.fft2(Z).

        Args:
            Z: data matrix, (N, C, H, W)

        Returns:
            The coding rate of Z, ().
        """
        N, C, H, W = Z.shape  # (N, C, H, W)
        Sigma_hat_data = F.Sigma_hat_translation_invariant(Z)  # (H, W, C, C)
        beta = C / self.eps_sq
        I = torch.eye(C).view(1, 1, C, C)  # (1, 1, C, C)
        Sigma_hat_distorted = I + beta * Sigma_hat_data  # (H, W, C, C)
        return torch.sum(F.logdet_complex(Sigma_hat_distorted)) / 2.0  # ()

    def R_per_class(self, Z, Pi):
        """
        Computes the coding rate of each class of Z as given by class information matrix Pi.
        NOTE: Only an accurate measure of the coding rate if the data is FFTed beforehand.
        You can do this with mcr2.functional.fft2(Z).

        Args:
            Z: data matrix, (N, C, H, W)
            Pi: class matrix, (N, K)

        Returns:
            The coding rate of each class of Z, (K, ).
        """
        N, C, H, W = Z.shape  # (N, C, H, W)
        Sigma_hat_data_per_class = F.Sigma_hat_per_class_translation_invariant(Z, Pi)  # (K, H, W, C, C)
        beta = C / self.eps_sq  # ()
        I = torch.eye(C).view(1, 1, 1, C, C)  # (1, 1, 1, C, C)
        Sigma_hat_distorted_per_class = I + beta * Sigma_hat_data_per_class  # (K, H, W, C, C)
        logdets_per_class = torch.sum(
            F.logdet_complex(Sigma_hat_distorted_per_class),  # (K, H, W)
            dim=(1, 2)
        )  # (K, )
        return logdets_per_class / 2.0  # (K, )

    def Rc(self, Z, Pi):
        """
        Computes the segmented coding rate of Z with respect to a class information matrix Pi.
        NOTE: Only an accurate measure of the coding rate if the data is FFTed beforehand.
        You can do this with mcr2.functional.fft2(Z).

        Args:
            Z: data matrix, (N, C, H, W)
            Pi: class information matrix, (N, K)

        Returns:
            The segmented coding rate of Z with respect to Pi, ().
        """
        N, K = Pi.shape  # (N, K)
        N_per_class = torch.sum(Pi, axis=0)  # (K, )
        gamma_per_class = N_per_class / N  # (K, )
        Rs_per_class = self.R_per_class(Z, Pi)  # (K, )
        return torch.sum(gamma_per_class * Rs_per_class)  # ()

    def DeltaR(self, Z, Pi):
        """
        Computes the coding rate reduction of Z with respect to a class information matrix Pi.
        NOTE: Only an accurate measure of the coding rate if the data is FFTed beforehand.
        You can do this with mcr2.functional.fft2(Z).

        Args:
            Z: data matrix, (N, C, H, W)
            Pi: class information matrix, (N, K)

        Returns:
            The coding rate reduction of Z with respect to Pi, ().
        """
        return super().DeltaR(Z, Pi)

    def DeltaR_distance(self, Z1, Z2):
        """
        Computes the DeltaR distance between the two data point sets Z1 and Z2.
        NOTE: Only an accurate measure of the distance if the data is FFTed beforehand.
        You can do this with mcr2.functional.fft2(Z).

        Args:
            Z1: data matrix, (N1, C, H, W)
            Z2: data matrix, (N2, C, H, W)

        Returns:
            The DeltaR distance between Z1 and Z2, ().
        """
        return super().DeltaR_distance(Z1, Z2)


__all__ = [
    "SupervisedVectorCodingRate",
    "SupervisedShiftInvariantCodingRate",
    "SupervisedTranslationInvariantCodingRate"
]



import torch


def logdet(H):
    """
    Takes the log determinant of a (possible batch of) real symmetric matrix.
    Args:
        H: A Hermitian matrix batch (*, D, D)

    Returns:
        The log determinant of H (*, )
    """
    ld = torch.slogdet(H)
    return ld[0] * ld[1]


def logdet_complex(H):
    """
    Takes the log determinant of a (possible batch of) complex Hermitian matrix.
    Args:
        H: A Hermitian matrix batch (*, D, D)

    Returns:
        The log determinant of H (*, )
    """
    ld = torch.slogdet(H)
    return (ld[0] * ld[1]).real


__all__ = ["logdet", "logdet_complex"]



import torch


def fft(Z):
    """
    Takes 1d FFT of the last dimension of Z.
    Args:
        Z: a data matrix (*, T)

    Returns:
        A tensor with the last dimension equal to FFT of Z (*, T)
    """
    return torch.fft.fft(Z, norm="ortho")


def ifft(Z):
    """
    Takes 1d IFFT of the last dimension of Z.
    Args:
        Z: an FFTed data matrix (*, T)

    Returns:
        A tensor with the last dimension equal to IFFT of Z (*, T)
    """
    return torch.fft.ifft(Z, norm="ortho")


def fft2(Z):
    """
    Takes 2d FFT of the last two dimensions of Z.
    Args:
        Z: a data matrix (*, H, W)

    Returns:
        A tensor with the last two dimensions equal to 2d FFT of Z (*, H, W)
    """
    return torch.fft.fft2(Z, norm="ortho")


def ifft2(Z):
    """
    Takes 2d IFFT of the last two dimensions of Z.
    Args:
        Z: an FFTed data matrix (*, H, W)

    Returns:
        A tensor with the last two dimensions equal to 2d IFFT of Z (*, H, W)
    """
    return torch.fft.ifft2(Z, norm="ortho")


__all__ = ["fft", "fft2", "ifft", "ifft2"]



import mcr2
import torch


class SupervisedMCR2Loss(torch.nn.Module):
    def __init__(self, eps_sq):
        super(SupervisedMCR2Loss, self).__init__()
        self.cr = self._get_coding_rate(eps_sq)

    def _get_coding_rate(self, eps_sq):
        raise NotImplementedError("Initialized abstract mcr2 class")

    def forward(self, Z, Pi):
        return -self.cr.DeltaR(Z, Pi)


class SupervisedVectorMCR2Loss(SupervisedMCR2Loss):
    def _get_coding_rate(self, eps_sq):
        return mcr2.coding_rate.SupervisedVectorCodingRate(eps_sq)


class SupervisedShiftInvariantMCR2Loss(SupervisedMCR2Loss):
    def _get_coding_rate(self, eps_sq):
        return mcr2.coding_rate.SupervisedShiftInvariantCodingRate(eps_sq)


class SupervisedTranslationInvariantMCR2Loss(SupervisedMCR2Loss):
    def _get_coding_rate(self, eps_sq):
        return mcr2.coding_rate.SupervisedTranslationInvariantCodingRate(eps_sq)


__all__ = ["SupervisedVectorMCR2Loss", "SupervisedShiftInvariantMCR2Loss", "SupervisedTranslationInvariantMCR2Loss"]



import torch


def normalize_frob(Z):  # (N, *)
    """
    Normalizes the tensor Z by dividing it by its Frobenius norm.
    Note: We use Frobenius normalization, not column-by-column normalization.
    This is for a variety of theoretical and empirical reasons.

    Args:
        Z: the data tensor (N, *)

    Returns:
        A normalized version of Z (N, *).
    """
    return Z / torch.linalg.norm(Z)


def normalize_vec(Z):  # (N, D)
    """
    Normalizes each row of the tensor Z by dividing it by its Frobenius (l2) norm.
    Args:
        Z: the data tensor (N, D)

    Returns:
        A normalized version of Z (N, D)
    """
    return torch.nn.functional.normalize(Z, dim=1)


def normalize_1d(Z):  # (N, C, T)
    """
    Normalizes each row of the tensor Z by dividing it by its Frobenius norm.
    Args:
        Z: the data tensor (N, C, T)

    Returns:
        A normalized version of Z (N, C, T)
    """
    N, C, T = Z.shape
    return torch.nn.functional.normalize(Z.reshape(N, C * T), dim=1).reshape(N, C, T)


def normalize_2d(Z):  # (N, C, H, W)
    """
    Normalizes each row of the tensor Z by dividing it by its Frobenius norm.
    Args:
        Z: the data tensor (N, C, H, W)

    Returns:
        A normalized version of Z (N, C, H, W)
    """
    N, C, H, W = Z.shape
    return torch.nn.functional.normalize(Z.reshape(N, C * H * W), dim=1).reshape(N, C, H, W)


__all__ = ["normalize_vec", "normalize_1d", "normalize_2d", "normalize_frob"]



import torch


def pi_to_y(Pi):
    """
    Turns a class membership matrix into a vector of class labels.
    Args:
        Pi: A one-hot encoded class membership matrix (N, K)

    Returns:
        A vector y of class labels (N, )
    """
    return torch.argmax(
        Pi,  # (N, K)
        dim=1
    )  # (N, )


def y_to_pi(y, K=-1):
    """
    Turns a vector of class labels y into a one-hot encoding class membership matrix Pi.
    Note: Assuming classes are indices {0, 1, ..., K - 1}. If K is not provided,
    picks K = max(y) + 1.

    Args:
        y: The vector of class labels.
        K: The number of classes, if provided. Optional since we can estimate it from given labels.

    Returns:
        A class membership matrix Pi (N, K)
    """
    return torch.nn.functional.one_hot(y, K)  # (N, K)


__all__ = ["pi_to_y", "y_to_pi"]
